/**
 * @ingroup     sc68_devel
 * @file        sc68.c
 * @author      Ben(jamin) Gerard <ben@sashipa.com>
 * @date        2001/12/05
 * @brief       sc68 command line player
 * @version     $Id: sc68.c,v 1.1.1.1 2001/12/24 20:50:11 ben Exp $
 */

/*
 *     sc68 - atari st and amiga music emulator - command line player
 *      Copyright (C) 1998-2001 Ben(jamin) Gerard <ben@sasghipa.com>
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2 of the License, or (at your
 *  option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

/** @defgroup SC68  sc68 command line player
 *
 *  sc68 - atari st and amiga music emulator - command line player
 *  Copyright (C) 1998-2001 Ben(jamin) Gerard <ben@sasghipa.com>
 *
 *  This program is free software.
 *
 *  usage
 *
 * options:
 *  --help         : Display this message and exit\n"
 *  --version      : Display sc68 version x.y.z and exit\n"
 *  --track=#      : Choose track to play [0=disk default]\n"
 *
 */

/* Standard Includes */
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <math.h>

/* generated config include */
#include "config.h"

/* sc68 includes */
#include "sc68app/SC68.h"
#include "file68/SC68error.h"
#include "file68/SC68utils.h"
#include "file68/SC68file.h"
#ifdef WITH_BROWSE68
# include "browse68/SC68browse.h"
#endif
#include "file68/SC68os_general.h"   
#include "sc68app/SC68os_interface.h"
#include "sc68app/SC68play.h"        
#include "sc68app/SC68delay.h"
#include "sc68app/SC68mixer.h"

/* 68000 emulator includes */
#include "emu68/emu68.h"
#include "emu68/ioplug68.h"
#include "io68/io68.h"

/* 512 Kb 68K memory buffer */
#define MEMORY (1<<19)

/* Browse info update when disk is loaded. */
#ifdef WITH_BROWSE68
# define UPDATE_BROWSE_INFO 1
#else
# define UPDATE_BROWSE_INFO 0
#endif

static SC68app_t app; /**< sc68 application context. */


/** Display version number. */
static int Version(void)
{
  puts("sc68 version " VERSION "\n");
  return 0;
}

/** Display usage message. */
static int Usage(void)
{
  puts(
    "sc68 - atari st and amiga music emulator\n"
    "Copyright (C) 1998-2001 Ben(jamin) Gerard <ben@sasghipa.com>\n"
    "\n"
    "This program is free software.\n"
    "\n"
    "usage : sc68 [options] input-sc68-file\n"
    "\n"
    "options:\n"
    "  --help         : Display this message and exit\n"
    "  --version      : Display sc68 version x.y.z and exit\n"
    "  --track=#      : Choose track to play [0=disk default]\n"
    "\n"
    );
  return 1;
}

/** Display to output debug statcked error messages.
 */
static void spool_error_message(void)
{
  char *s;

  if (s = SC68error_get(), s) {
    fprintf(stderr, "Stacked Error Message:\n");
	  do {
	    fprintf(stderr, "%s\n", s);
	  } while (s = SC68error_get(), s != NULL);
  } else {
    fprintf(stderr, "No stacked error message\n");
  }
}
                          

/** Initialize 68000 emulator.
 */
static int init_68K(reg68_t * r, u32 sz)
{
  if (EMU68_init(sz) < 0) {
    return SC68error_add(EMU68error_get());
  }
  *r = reg68;
  r->sr = 0x2000;
  r->a[7] = sz - 4;
  if (YM_init() < 0)
    return SC68error_add("Yamaha 2149 emulator init failed");
  if (MW_init() < 0)
    return SC68error_add("Microwire emulator init failed");
  if (MFP_init() < 0)
    return SC68error_add("MFP emulator init failed");
  if (PL_init() < 0)
    return SC68error_add("Paula emulator init failed");
  return 0;
}

/** Set sc68 application to default values. */
static void set_sc68app_default(SC68app_t * app)
{
  memset(app, 0, sizeof(*app));
  app->cur_track = -1;
  SC68config_default(&app->config);
}

static int sc68_init(char *execname)
{
  int r, frq;

  SC68os_pdebug("\nINIT sc68 player\n");

  /* check whether EMU68 is compiled without debug option */
  if (EMU68_sizeof_reg68() != sizeof(reg68_t)) {
    SC68error_add
      ("Internal : Bad EMU68 version : should not define SC68DEBUG ...");
    return -1;
  }

  /* Set default configuration */
  set_sc68app_default(&app);

  /* Get exec path : for resource */
  if (SC68os_get_exec_path(execname) < 0) {
    return -2;
  }

  /* Get the config file OR create if it doesn't exist */
  SC68os_pdebug("Read config file ...\n");
  if (SC68config_load(&app.config,
                      SC68os_resource_name(SC68rsc_config,
                                           SC68_DEFAULT_CONFIG_FILE)) < 0) {
    SC68error_get();
    SC68error_get();
    SC68config_save(&app.config,
                    SC68os_resource_name(SC68rsc_config,
                                         SC68_DEFAULT_CONFIG_FILE));
  }
  SC68os_pdebug("OK\n");

  /*SC68config_display(&app.config);*/

  /* Open application ( OS dependant code ) */
  r = SC68os_interface_open(execname, &app);

  /* Open error */
  if (r < 0) {
    return -3;
  }

  /* Copy config parameters to app */
  SC68app_configure(&app);

  /* Delay init : from config file parameters ( SC68app_configure() ) */
  app.delay.buffer_size = SC68delay_init(app.delay.buffer_size);
  if (app.delay.buffer_size < 0)
    return -4;

  /* Copy config parameters to app */
  SC68app_configure(&app);

  /* Delay init : from config file parameters ( SC68app_configure() ) */
  app.delay.buffer_size = SC68delay_init(app.delay.buffer_size);
  if (app.delay.buffer_size < 0)
    return -5;

  frq = (int) app.mix.real_frq;

  /* First instance : Do mucho initialization */
  if (r > 0) {
#ifdef WITH_BROWSE68
    /* Browse Info Data Base init */
    if (SC68browse_init(SC68_DEFAULT_BIDB_FILE) < 0) {
      return -6;
    }
    SC68os_pdebug("Browse init : OK\n");
#endif
    /* 68K & IO-Emulators init */
    if (init_68K(&app.reg68, MEMORY)) {
      return -7;
    }
    SC68os_pdebug("68K emulator init : OK\n");

    /* IO replay frequency init */
    app.mix.real_frq = 44100;
    YM_set_replay_frq(app.mix.real_frq);
    MW_set_replay_frq((int) app.mix.real_frq);
    PL_set_replay_frq((int) app.mix.real_frq);
    app.mix.buf = YM_get_buffer();
  }
  return 0;
}

static int PlayLoop(FILE *out)
{
  int done = 0;
  
  while (!done) {
    unsigned int sign = SC68MIXER_CHANGE_SIGN;
    unsigned int cycle_this_pass = app.mix.cycleperpass;
    
    if (!app.mix.buflen) {
      EMU68_level_and_interrupt(cycle_this_pass);

      /* YM ??? */
      if (app.cur_mus->flags.ym) {
        app.mix.buflen = YM_mix(cycle_this_pass);
      } else {
        app.mix.buflen = app.mix.stdbuflen;
      }

      app.mix.buf = YM_get_buffer();

      /* MW ??? */
      if (app.cur_mus->flags.ste) {
        MW_mix(app.mix.buf, app.reg68.mem, app.mix.buflen);
        if (sign) {
          SC68mixer_stereo_16_LR(app.mix.buf, app.mix.buf,
				 app.mix.buflen, sign);
        }
      }
      /* Paula ??? */
      else if (app.cur_mus->flags.amiga) {
        PL_mix(app.mix.buf, app.reg68.mem, app.reg68.mem + app.reg68.memsz,
               app.mix.buflen);

        SC68mixer_stereo_16_LR(app.mix.buf,app.mix.buf, app.mix.buflen,
                               SC68MIXER_CHANGE_SIGN);

        SC68mixer_blend_LR(app.mix.buf, app.mix.buf, app.mix.buflen,
                           app.mix.amiga_blend , sign);
      }
      /* Do stereo my self */
      else {
        SC68mixer_dup_L_to_R(app.mix.buf, app.mix.buf, app.mix.buflen, sign);
      }

      if (app.delay.onoff) {
        /* Apply delay ... */
        int delay;
        const int shift = 0;

        if (app.delay.time < 0)
          delay = (app.mix.buflen << shift) / -app.delay.time;
        else
          delay = (app.mix.buflen * app.delay.time) >> shift;
        if (!app.delay.lr) {
          /* Delay ring */
          SC68delay (app.mix.buf, app.mix.buflen, delay,
		     app.delay.strength, sign);
        } else {
          /* L/R spacializer delay */
          SC68delay_LR(app.mix.buf, app.mix.buflen, delay);
	}
      }

      /* Advance time */
      SC68app_advance_time(&app);

      /* Reach end of track */
      if (app.time.elapsed > app.time.track) {
        if(app.cur_track+1 >= app.cur_disk->nb_six) {
          done = 1;
        } else {
          SC68track(&app, app.cur_track+1);
	}
      }
    }

    if (fwrite (app.mix.buf, app.mix.buflen << 2, 1, out) != 1) {
      SC68error_add(strerror(errno));
      SC68error_add("Audio data write error (see previous message)");
      done = -1;
    }
    app.mix.buflen = 0;
  }
  SC68eject(&app);
  return done < 0 ? done : 0;
}


static int IsIntParam(const char *parm, const char *what, int *res)
{
  if (strstr(parm, what) == parm) {
    int len = strlen(what);
    if (isdigit(parm[len])) {
      *res = atoi(parm+len);
      return 1;
    }
  }
  return 0;
}

int main(int na, char **a)
{
  int i;
  int help = 0;
  char *fname = 0;
  int track=0;
  disk68_t *disk;

  for (i=1; i<na; ++i) {
    if (!strcmp(a[i],"--help")) {
      help = 1;
      break;
    } else if (!strcmp(a[i],"--version")) {
      return Version();
    }
  }

  if (na<2 || help) {
    return Usage();
  }

  for (i=1; i<na; ++i) {
    if (!IsIntParam(a[i],"--track=", &track)) {
      if (fname) {
	fprintf(stderr, "Invalid parameters \"%s\"\n", a[i]);
	return 2;
      } else {
	fname = a[i];
      }
    }
  }

  if (!fname) {
    fprintf (stderr, "Missing input file. try --help\n");
    return 3;
  }
  
  fname = a[1];
  
  /* Verify sc68 file. */
  if (SC68file_verify(fname) < 0) {
    fprintf(stderr, "Not a valid sc68 file \"%s\"\n", fname);
    return 4;
  }
  
  /* Init sc68 application context and 68K emulator */
  if (sc68_init(a[0]) < 0) {
    spool_error_message();
    return 5;
  }

  /* Load disk */
  disk = SC68app_load_diskfile(fname, UPDATE_BROWSE_INFO);
  if (!disk) {
    spool_error_message();
    return 6;
  }

  /* Set track number : command line is prior to config force-track */
  if (track) {
    app.force_track = track;
  }
  /* Load disk in application context : this set the track too. */
  if (SC68load(&app, disk) < 0) {
    spool_error_message();
    return 7;
  }

  /* Play current track of current disk. */
  if (SC68play(&app) < 0) {
    spool_error_message();
    return 8;
  }

  /** @todo  Set out stream mode to binary. */

  /* Loop */
  if (PlayLoop(stdout) < 0) {
    spool_error_message();
    return 9;
  }

  return 0;
}
